from dlchord2.note import Note
from dlchord2.parser.quality_parser import QualityParser


class Quality(object):
    def __init__(self, quality_text):
        self.__raw_quality_text = quality_text
        parser = QualityParser()
        self.__quality_data = parser.parse(quality_text)

        self.__tension_notes = [Note.create_from_tension(tension)
                                for tension in self.__quality_data.tensions]
        self.__tension_parentheses_notes = [Note.create_from_tension(tension)
                                            for tension in self.__quality_data.tensions_parentheses]
        self.__add_tension_notes = [Note.create_from_tension(tension)
                                    for tension in self.__quality_data.add_tensions]

    def __str__(self):
        return self.__raw_quality_text

    def __unicode__(self):
        return self.__raw_quality_text

    def __repr__(self):
        return "<Quality: {}>".format(self.__raw_quality_text)

    def __eq__(self, other):
        if not isinstance(other, Quality):
            raise TypeError("{} オブジェクトとQualityオブジェクトを比較できません".format(type(other)))

        if [note.note_index for note in self.get_notes()] == [note.note_index for note in  other.get_notes()]:
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def raw_chord_text(self):
        return self.__raw_chord_text

    @property
    def raw_quality_text(self):
        """
        生のクオリティテキストを取得します。
        :return: 生のクオリティテキスト
        :rtype: str
        """
        return self.__raw_quality_text

    @property
    def qualities(self):
        """
        クオリティのリストを取得します。
        :return: クオリティのリスト
        :rtype: list[str]
        """
        return self.__quality_data.qualities

    @property
    def tension_parent_notes(self):
        """
        かっこ内のテンションノートのリストを取得します。
        :return: かっこ内のテンションノートのリスト
        :rtype: list[Note]
        """
        return self.__tension_parentheses_notes

    @property
    def tension_notes(self):
        """
        かっこ外のテンションノートのリストを取得します。
        このテンションノートリストは、テンションのヒエラルキー構造処理(最上位のテンション以下のテンションを作成する処理)は行われません。
        :return: かっこ外のテンションノートのリスト
        :rtype: list[Note]
        """
        return self.__tension_notes

    @property
    def add_tension_notes(self):
        """
        アド・テンションノートのリストを取得します。
        :return: アド・テンションノートのリスト
        :rtype: list[Note]
        """
        return self.__add_tension_notes

    def __get_quality_value(self):
        """
        コードクオリティの構成音に与える変化量を取得します。
        :return: コードクオリティの構成音に与える変化量
        :rtype: list[int]
        """
        quality_value = [0, 0, 0, 0]
        for quality in self.qualities:
            if quality == "M":
                quality_value[3] += 1
            elif quality == "m":
                quality_value[1] -= 1
            elif quality == "dim":
                quality_value[1] -= 1
                quality_value[2] -= 1
                quality_value[3] -= 1
            elif quality == "aug":
                quality_value[2] += 1

        return quality_value

    def get_hierarchy_tensions_all(self):
        """
        テンションのヒエラルキー構造処理をし、すべてのテンションを取得します。
        :return: かっこ内、かっこ外、すべてのテンションのリスト
        :rtype: list[str]
        """
        tensions = []

        for tension in self.__quality_data.tensions:
            # 階層構造のないテンション
            if tension == "4" or tension == "6" or tension == "2":
                tensions.append(tension)

            # 最上位
            elif tension == "13":
                tensions.append("13")
                tensions.append("11")
                tensions.append("9")
            elif tension == "11":
                tensions.append("11")
                tensions.append("9")
            elif tension == "9":
                tensions.append("9")
            else:
                tensions.append(tension)

        tensions.extend(self.__quality_data.tensions_parentheses)
        tensions.extend(self.__quality_data.add_tensions)

        return list(set(tensions))

    def get_notes(self):
        """
        このクオリティから表されるノートのリストを取得します。
        :return: ノートのリスト
        :rtype: list[Note]
        """

        quality_notes = [Note("C"), Note("E"), Note("G"), Note("Bb")]

        # 7thの無効化
        quality_notes[3].disabled = True

        # クオリティによるノートの変化
        quality_values = self.__get_quality_value()
        for i, value in enumerate(quality_values):
            quality_notes[i] = quality_notes[i].transposed(value)

        # 直接変更されるクオリティによる変化
        qualities = self.qualities
        for qua in qualities:
            if qua == "sus":
                quality_notes[1].disabled = True

        # テンションの追加
        tensions = self.get_hierarchy_tensions_all()
        for tension in tensions:
            # -5や#5は既存のノートを変化させる
            if tension == "-5" or tension == "b5":
                quality_notes[2] = quality_notes[2].transposed(-1)
                continue
            elif tension == "+5" or tension == "#5":
                quality_notes[2] = quality_notes[2].transposed(1)
                continue
            elif tension == "5":
                quality_notes[1].disabled = True
                continue
            elif tension == "7":
                quality_notes[3].disabled = False
                continue

            tension_note = Note.create_from_tension(tension)
            quality_notes.append(tension_note)

        # 無効のノートを削除
        notes = [note for note in quality_notes if not note.disabled]
        return notes



from dlchord2 import const
from dlchord2.parser.chord_parser import ChordParser


def test_parse():
    parser = ChordParser()
    chord_data = parser.parse("B#/C")
    for quality_name, notes in const.CHORD_MAP.items():
        for root_text in const.TONE_TO_INDEX.keys():
            for bass_text in const.TONE_TO_INDEX.keys():
                chord_name = root_text + quality_name + "/" + bass_text
                parser = ChordParser()
                chord_data = parser.parse(chord_name)
                assert root_text == chord_data.root_text
                assert bass_text == chord_data.bass_text
                assert quality_name == chord_data.quality_text



from dlchord2 import const
from dlchord2.chord import Chord
from dlchord2.utils import match_quality, notes_to_chords, note_indexes_to_chords


def test_match_quality():
    for quality_name, notes in const.CHORD_MAP.items():
        is_match, quality = match_quality(notes)
        assert quality_name == quality


def test_notes_to_chords():
    tones = const.TONE_TO_INDEX.keys()
    qualities = const.CHORD_MAP.items()
    for quality_name, notes in qualities:
        for root_text in tones:
            for bass_text in tones:
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)

                chord_list = notes_to_chords(chord.get_notes())
                assert chord_list[0] == chord


def test_note_indexes_to_chords():
    tones = const.TONE_TO_INDEX.keys()
    qualities = const.CHORD_MAP.items()
    for quality_name, notes in qualities:
        for root_text in tones:
            for bass_text in tones:
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)

                chord_list = note_indexes_to_chords([note.note_index for note in chord.get_notes()])
                assert chord_list[0] == chord



import re

from dlchord2 import const
from dlchord2.exceptions.quality_exceptions import QualityParseError


class QualityParseData(object):
    def __init__(self, qualities, tensions, tensions_parentheses, add_tensions):
        self.__qualities = qualities
        self.__tensions = tensions
        self.__tensions_parentheses = tensions_parentheses
        self.__add_tensions = add_tensions

    @property
    def qualities(self):
        """
        テンションを含まないクオリティ部分のみのテキストリストを取得します。
        :return: クオリティのテキストリスト
        :rtype: list[str]
        """
        return self.__qualities

    @property
    def tensions(self):
        """
        かっこ外のテンションのリストを取得します。
        :return: かっこ外のテンションのリスト
        :rtype: list[str]
        """
        return self.__tensions

    @property
    def tensions_parentheses(self):
        """
        かっこ内のテンションのリストを取得します。
        :return: かっこ内のテンションのリスト
        :rtype: list[str]
        """
        return self.__tensions_parentheses

    @property
    def add_tensions(self):
        """
        アド・テンションを取得します。
        :return: アド・テンション
        :rtype: list[Note]
        """
        return self.__add_tensions


class QualityParser(object):
    """
    コードクオリティを解析するクラス
    """

    def __find_tension_parentheses(self, quality_text):
        """
        かっこ内のテンションノートを探します。
        :param quality_text: コードクオリティのテキスト
        :type quality_text: str
        :return: かっこ内のテンションのリスト
        :rtype: list[str]
        """
        match = re.search(r"(\()(.*?)(\))", quality_text)
        if match is None:
            return []

        if len(match.groups()) < 3:
            return []

        tensions_raw_text = match.group(2)
        tensions_text = tensions_raw_text.split(",")
        tensions = []

        for tension_text in tensions_text:
            tension = tension_text.strip()

            if tension not in const.TENSION_TO_INDEX:
                raise QualityParseError("かっこ内のテンション {} が不明です".format(tension))

            tensions.append(tension)
        return tensions

    def __find_tension(self, quality_text):
        """
        かっこ外のテンションノートを探します。
        :param quality_text: コードクオリティのテキスト
        :type quality_text: str
        :return: かっこ外のテンションのリスト
        :rtype: list[str]
        """
        exclude_parentheses = re.sub(r"(\()(.*?)(\))", "", quality_text)

        if "add" in exclude_parentheses:
            return ""

        match_tensions = re.findall("9|11|13|[-+b#]5|2|4|6|(?<![-+b#])5", exclude_parentheses)

        for tension in match_tensions:
            if tension not in const.TENSION_TO_INDEX:
                raise QualityParseError("かっこ外のテンション {} が不明です".format(tension))

        return match_tensions

    def __find_add_tension(self, quality_text):
        """
        アド・テンションを取得します。
        :param quality_text: コードクオリティのテキスト
        :type quality_text: str
        :return: アド・テンションノートのリスト
        :rtype: list[str]
        """
        match = re.findall(r"add9|add11|add4|add2", quality_text)

        add_tension_list = []
        for tension in match:
            add_tension_list.append(tension.replace("add", ""))

        return add_tension_list

    def __find_quality(self, quality_text):
        """
        テンションを含まないクオリティを取得します。
        :param quality_text: コードクオリティのテキスト
        :type quality_text: str
        :return: クオリティのリスト
        :rtype: list[str]
        """
        qualities = []
        match = re.findall(r"aug|dim|sus", quality_text)
        if match:
            qualities.extend(match)
            quality_text = re.sub(r"aug|dim|sus", "", quality_text)

        match = re.findall(r"m|M", quality_text)
        if match:
            qualities.extend(match)
            quality_text = re.sub(r"m|M", "", quality_text)

        return qualities

    def __exists_seventh(self, quality_text, tensions):
        if "7" in quality_text:
            return True

        for tension in tensions:
            if tension == "13":
                return True
            elif tension == "11":
                return True
            elif tension == "9" and "6" not in tensions:
                return True

        return False

    def parse(self, quality_text):
        tensions_parentheses = self.__find_tension_parentheses(quality_text)
        tensions = self.__find_tension(quality_text)
        add_tensions = self.__find_add_tension(quality_text)
        qualities = self.__find_quality(quality_text)

        exists_seventh = self.__exists_seventh(quality_text, tensions)
        if exists_seventh:
            tensions.append("7")

        quality_data = QualityParseData(qualities, tensions, tensions_parentheses, add_tensions)
        return quality_data



from dlchord2.note import Note
from dlchord2.parser.accidentals_parser import AccidentalsType
from dlchord2.parser.chord_parser import ChordParser
from dlchord2.quality import Quality
from dlchord2 import utils


class Chord(object):
    def __init__(self, chord_text, scale_text="C"):
        self.__raw_chord_text = chord_text
        self.__scale_text = scale_text

        parser = ChordParser()
        self.__chord_data = parser.parse(chord_text)
        self.__root = Note(self.__chord_data.root_text, scale_text=scale_text)
        self.__bass = Note(self.__chord_data.bass_text, scale_text=scale_text)
        self.__quality = Quality(self.__chord_data.quality_text)

    def __str__(self):
        return self.raw_chord_text

    def __unicode__(self):
        return self.raw_chord_text

    def __repr__(self):
        return "<Chord: {}>".format(self.raw_chord_text)

    def __eq__(self, other):
        if not isinstance(other, Chord):
            raise TypeError("{} オブジェクトとChordオブジェクトを比較できません".format(type(other)))

        note_indexes = {note.note_index for note in self.get_notes()}
        other_note_indexes = {note.note_index for note in self.get_notes()}
        if note_indexes == other_note_indexes:
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def raw_chord_text(self):
        return self.__raw_chord_text

    @property
    def root(self):
        """
        ルート音を取得します。
        :return: ルート音
        :rtype: Note
        """
        return self.__root

    @property
    def bass(self):
        """
        ベース音を取得します。
        :return: ベース音
        :rtype: Note
        """
        return self.__bass

    @property
    def quality(self):
        """
        コードクオリティを取得します。
        :return: コードクオリティ
        :rtype: Quality
        """
        return self.__quality

    @property
    def is_on_chord(self):
        """
        オンコードかを取得します。
        :return: オンコードの場合はTrue,それ以外の場合はFalse
        :rtype: bool
        """
        return self.root != self.bass

    def transposed(self, steps, scale_text=None):
        """
        転調したコードを返します。
        scale_textがNoneの場合は、現在のコードに設定されているスケールを使用します。
        :param steps: 転調する数
        :type steps: int
        :param scale_text: 転調後のコードに適用するスケール
        :type scale_text: str
        :return: 転調後のコード
        :rtype: Chord
        """
        if self.is_on_chord:
            return Chord(self.root.transposed(steps, scale_text).normed_note_text +
                         self.quality.raw_quality_text +
                         "/" + self.bass.transposed(steps, scale_text).normed_note_text)
        else:
            return Chord(self.root.transposed(steps, scale_text).normed_note_text +
                         self.quality.raw_quality_text)

    def get_notes(self, sparse=False):
        """
        コードのノートリストを取得します。
        :param sparse: ノートリストをスパースにするか
        :return: コードのノートリスト
        :rtype: list[Note]
        """
        relative_notes = self.quality.get_notes()
        notes = [note.transposed(self.root.note_index) for note in relative_notes]

        if self.is_on_chord:
            notes.insert(0, self.bass)

        if sparse:
            # 非構成音は0
            sparse_notes = [0 for i in range(12)]

            # 構成音は1
            for note in notes:
                sparse_notes[note] = 1

            # ベース音は2
            sparse_notes[notes[0]] = 2

            notes = sparse_notes

        return notes

    def reconfigured(self):
        """
        コードを再構成します。
        :return: 再構成されたコード
        :rtype: Chord
        """

        rec_chord = utils.notes_to_chords(self.get_notes())[0]

        if rec_chord.is_on_chord:
            bass_chord = rec_chord.bass.transposed(0, rec_chord.root.raw_note_text)
            rec_chord = Chord(rec_chord.root.raw_note_text +
                              rec_chord.quality.raw_quality_text +
                              "/" + bass_chord.normed_note_text, scale_text=self.__scale_text)
        else:
            rec_chord = Chord(rec_chord.root.raw_note_text +
                              rec_chord.quality.raw_quality_text, scale_text=self.__scale_text)

        return rec_chord



from dlchord2 import const
from dlchord2 import chord as chd
from dlchord2.note import Note


def __chord_sort_func(chord):
    eval_score = 0
    if Note.create_from_tension("5") in chord.quality.tension_notes:
        eval_score -= 1

    if chord.bass in chord.quality.add_tension_notes:
        eval_score -= 1

    if chord.is_on_chord:
        eval_score -= 1

    seventh_index = 10
    if chord.bass.note_index == (chord.root.note_index + seventh_index) % 12:
        if Note.create_from_tension("7") in chord.quality.tension_notes:
            eval_score += 1

            if "aug" in chord.quality.qualities:
                eval_score -= 1

    return eval_score


def match_quality(notes, omit_five_note=False, omit_third_note=False):
    """
    ノートのリストから一致するクオリティを検索します。
    :param notes: ノートのリスト
    :type notes: list[int]
    :param omit_five_note: 5度のノートを省略して検索するかどうか
    :type omit_five_note: bool
    :param omit_third_note: 3度のノートを省略して検索するかどうか
    :type omit_third_note: bool
    :return: マッチしたらTrueとクオリティを返します。それ以外の場合はFalseと""を返します。
    :rtype: tuple(bool, str)
    """
    relative_notes = []
    for i in range(len(notes)):
        relative_notes.append((notes[i] - notes[0]) % 12)

    relative_notes = set(relative_notes)

    for quality_name, notes_origin in const.CHORD_MAP.items():
        notes_origin = set(notes_origin)
        if relative_notes == notes_origin:
            return True, quality_name

        omitted_notes = relative_notes.copy()
        # 5度の音を省略して検索
        if omit_five_note:
            five_note_index = 7
            if five_note_index in omitted_notes:
                omitted_notes.remove(five_note_index)
            if five_note_index in notes_origin:
                notes_origin.remove(five_note_index)

            if omitted_notes == notes_origin:
                return True, quality_name

        # 3度の音を省略して検索
        if omit_third_note:
            third_note_index = 4
            if third_note_index in omitted_notes:
                omitted_notes.remove(third_note_index)
            if third_note_index in notes_origin:
                notes_origin.remove(third_note_index)

            if omitted_notes == notes_origin:
                return True, quality_name

    return False, ""


def notes_to_chords(notes, sort_func=__chord_sort_func):
    """
    ノーツリストから該当するコードの候補のリストを取得します。
    :param notes: ノーツリスト
    :type notes: list[Note]
    :param sort_func: コードをソートするときに使用する関数
    :type sort_func: function
    :return: コードの候補のリスト
    :rtype: list[Chord]
    """
    note_indexes = [note.note_index for note in notes]
    return note_indexes_to_chords(note_indexes, sort_func=sort_func)


def note_indexes_to_chords(note_indexes, sort_func=__chord_sort_func):
    """
    ノーツリストから該当するコードの候補のリストを取得します。
    :param note_indexes: ノーツリスト
    :type note_indexes: list[int]
    :param sort_func: コードをソートするときに使用する関数
    :type sort_func: function
    :return: コードの候補のリスト
    :rtype: list[Chord]
    """

    # 重複ノーツは不要
    note_indexes = sorted(set(note_indexes), key=note_indexes.index)

    shifted_notes = note_indexes[1:]

    chord_list = []
    bass_note = note_indexes[0]
    # ベースを含まない転回系
    for i in range(len(shifted_notes)):
        is_match, quality = match_quality(shifted_notes)
        if is_match:
            root_note = shifted_notes[0]

            on_chord = ""
            if bass_note != root_note:
                on_chord = "/" + Note.create_from_index_note(bass_note).normed_note_text

            chord = chd.Chord(Note.create_from_index_note(root_note).normed_note_text + quality + on_chord)
            chord_list.append(chord)

        note = shifted_notes.pop(0)
        shifted_notes.append(note)

    # ベースを含む転回系
    shifted_notes.insert(0, bass_note)
    for i in range(len(shifted_notes)):
        is_match, quality = match_quality(shifted_notes)
        if is_match:
            root_note = shifted_notes[0]

            on_chord = ""
            if bass_note != root_note:
                on_chord = "/" + Note.create_from_index_note(bass_note).normed_note_text

            chord = chd.Chord(Note.create_from_index_note(root_note).normed_note_text + quality + on_chord)
            chord_list.append(chord)

        note = shifted_notes.pop(0)
        shifted_notes.append(note)

    chord_list = sorted(chord_list, key=sort_func, reverse=True)
    return chord_list



from dlchord2 import const
from dlchord2.parser.note_parser import NoteParser
from dlchord2.scale import Scale


class Note(object):
    def __init__(self, note_text, scale_text="C"):
        self.__raw_note_text = note_text
        self.__is_disabled = False

        parser = NoteParser()
        self.__note_data = parser.parse(self.__raw_note_text)

        if scale_text is None:
            scale_text = "C"

        self.__scale = Scale(scale_text)

    def __eq__(self, other):
        if not isinstance(other, Note):
            raise TypeError("{} オブジェクトとNoteオブジェクトを比較できません。".format(type(other)))

        if self.note_index == other.note_index:
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __unicode__(self):
        return self.raw_note_text

    def __str__(self):
        return self.raw_note_text

    def __repr__(self):
        return "<Note: {}>".format(self.raw_note_text)

    @classmethod
    def create_from_index_note(cls, index):
        """
        インデックスからノートを作成します。
        :param index: ノートの音階のインデックス
        :type index: int
        :return: ノート
        :rtype: Note
        """
        if index < 0 or index > 11:
            raise ValueError("テンションノートインデックスに指定できるのは、0から11までの数値です")

        note_text = const.C_MAJOR_SCALE[index]
        note = cls(note_text)
        return note

    @classmethod
    def create_from_tension(cls, tension_text):
        """
        テンションノートテキストからノートを作成します
        :param tension_text: テンションノートテキスト
        :type tension_text: str
        :return: テンションノート
        :rtype: Note
        """

        if tension_text not in const.TENSION_TO_INDEX:
            raise ValueError("解析できない不明なテンションノート: {} が渡されました".format(tension_text))

        tension_note = cls.create_from_index_note(const.TENSION_TO_INDEX[tension_text])
        return tension_note

    def transposed(self, steps, scale_text=None):
        """
        ノートを転調します。
        このノートを直接変更はせずに、転調したノートを新たに生成し、返します。
        scale_textがNoneの場合は、現在のノートに設定されているスケールを使用します。
        :param steps: 転調する数
        :type steps: int
        :param scale_text: 転調後のノートに適用するスケール
        :type scale_text: str
        :return: 転調後のノート
        :rtype: Note

        """

        transposed_note_index = (self.note_index + steps) % 12
        if scale_text is None:
            scale_text = self.__scale.scale_text

        transposed_note = Note(self.__scale.get_note_text_absolute(transposed_note_index), scale_text)
        transposed_note.disabled = self.disabled
        return transposed_note

    @property
    def disabled(self):
        """
        ノートが無効になっているかを取得します。
        :return: ノートが無効になっているか
        :rtype: bool
        """
        return self.__is_disabled

    @disabled.setter
    def disabled(self, disabled):
        """
        ノートを無効にするかを設定します。
        :param disabled: ノートを無効にする場合はTrue, それ以外はFalse
        :type disabled: bool
        """
        self.__is_disabled = disabled

    @property
    def note_index(self):
        """
        ノートのインデックスを取得します。
        ノートのインデックスは、Cを0として、Bを11とする音階のインデックスです。
        :return: ノートのインデックス
        :rtype: int
        """
        return self.__note_data.note_index

    @property
    def raw_note_text(self):
        """
        正規化されていないノートのテキストを取得します。
        :return: 正規化されていないノートのテキスト
        :rtype: str
        """
        return self.__raw_note_text

    @property
    def normed_note_text(self):
        """
        このノートに設定されているスケールで正規化したノートのテキストを取得します。
        :return: スケールで正規化したノートのテキスト
        :rtype: str
        """
        return self.scale.get_note_text_absolute(self.note_index)

    @property
    def accidentals(self):
        """
        調号を取得します。
        :return: 調号
        :rtype: Accidentals
        """
        return self.__note_data.accidentals

    @property
    def scale(self):
        """
        このノートに設定されているスケールを取得します。
        :return: このノートに設定されているスケール
        :rtype: Scale
        """
        return self.__scale




import copy

from dlchord2 import const
from dlchord2.chord import Chord
from dlchord2.note import Note


def test_root():
    for quality_name, notes in const.CHORD_MAP.items():
        for root_text in const.TONE_TO_INDEX.keys():
            for bass_text in const.TONE_TO_INDEX.keys():
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)
                assert const.TONE_TO_INDEX[root_text] == chord.root.note_index


def test_bass():
    for quality_name, notes in const.CHORD_MAP.items():
        for root_text in const.TONE_TO_INDEX.keys():
            for bass_text in const.TONE_TO_INDEX.keys():
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)
                assert const.TONE_TO_INDEX[bass_text] == chord.bass.note_index


def test_quality():
    for quality_name, notes in const.CHORD_MAP.items():
        for root_text in const.TONE_TO_INDEX.keys():
            for bass_text in const.TONE_TO_INDEX.keys():
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)
                assert quality_name == chord.quality.raw_quality_text


def test_is_on_chord():
    for quality_name, notes in const.CHORD_MAP.items():
        for root_text in const.TONE_TO_INDEX.keys():
            for bass_text in const.TONE_TO_INDEX.keys():
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)
                is_on_chord = Note(root_text) != Note(bass_text)
                assert is_on_chord == chord.is_on_chord

    for quality_name, notes in const.CHORD_MAP.items():
        for root_text in const.TONE_TO_INDEX.keys():
            chord_name = root_text + quality_name
            chord = Chord(chord_name)
            assert not chord.is_on_chord


def test_transposed():
    tones = const.TONE_TO_INDEX.keys()
    qualities = const.CHORD_MAP.items()
    for quality_name, notes in qualities:
        for root_text in tones:
            for bass_text in tones:
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)

                for i in range(1):
                    transposed_chord = chord.transposed(i)
                    transposed_chord_name = (Note(root_text).transposed(i).raw_note_text +
                                             quality_name + "/" +
                                             Note(bass_text).transposed(i).raw_note_text)
                    transposed_chord_origin = Chord(transposed_chord_name)

                    assert transposed_chord_origin == transposed_chord


def test_get_notes():
    tones = const.TONE_TO_INDEX.keys()
    qualities = const.CHORD_MAP.items()
    for quality_name, notes in qualities:
        for root_text in tones:
            for bass_text in tones:
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)

                note_indexes = {note.note_index for note in chord.get_notes()}
                notes_origin = list(copy.copy(notes))

                for i in range(len(notes_origin)):
                    notes_origin[i] += Note(root_text).note_index
                    notes_origin[i] %= 12

                if chord.is_on_chord:
                    notes_origin.insert(0, Note(bass_text).note_index)

                assert set(notes_origin) == note_indexes


def test_reconfigured():
    tones = const.TONE_TO_INDEX.keys()
    qualities = const.CHORD_MAP.items()
    for quality_name, notes in qualities:
        for root_text in tones:
            for bass_text in tones:
                chord_name = root_text + quality_name + "/" + bass_text
                chord = Chord(chord_name)
                rec_chord = chord.reconfigured()
                assert chord == rec_chord



from enum import Enum

from dlchord2.const import ACCIDENTALS_SHARP, ACCIDENTALS_FLAT
from dlchord2.exceptions.accidentals_exceptions import AccidentalsParseError


class AccidentalsType(Enum):
    """
    調号の種類を表す列挙体
    """
    NONE = 0
    SHARP = 1
    FLAT = 2


class AccidentalsParseData(object):
    """
    調号を解析したデータを格納するクラス
    """

    def __init__(self, accidentals, accidentals_type, transpose_num):
        self._accidentals = accidentals
        self._accidentals_type = accidentals_type
        self._transpose_num = transpose_num

    @property
    def accidentals(self):
        """
        調号のテキストを取得します。
        :return: 生の調号のテキスト
        :rtype: str
        """
        return self._accidentals

    @property
    def accidentals_type(self):
        """
        調号の種類を取得します。
        :return: 調号の種類
        :rtype: AccidentalsType
        """
        return self._accidentals_type

    @property
    def transpose_num(self):
        """
        調号の変化量を取得します。
        :return: 調号の変化量
        :rtype: int
        """
        return self._transpose_num


class AccidentalsParser(object):
    """
    調号を解析するクラス
    """

    def parse(self, accidentals_text):
        """
        調号を解析します。
        :param accidentals_text: ルート音を含まない調号テキスト
        :type accidentals_text: str
        :return: 調号解析データ
        :rtype: AccidentalsParseData
        """

        sharp_num = accidentals_text.count(ACCIDENTALS_SHARP)
        flat_num = accidentals_text.count(ACCIDENTALS_FLAT)

        if sharp_num > 0 and flat_num > 0:
            raise AccidentalsParseError("異なる調号は重複して存在することはできません。")

        accidentals_type = AccidentalsType.NONE
        trans_num = 0
        if sharp_num > 0:
            accidentals_type = AccidentalsType.SHARP
            trans_num = sharp_num

        elif flat_num > 0:
            accidentals_type = AccidentalsType.FLAT
            trans_num = -flat_num

        accidentals_parse_data = AccidentalsParseData(
            accidentals_text,
            accidentals_type,
            trans_num)
        return accidentals_parse_data



from dlchord2 import const
from dlchord2.accidentals import Accidentals
from dlchord2.parser.note_parser import NoteParser


def test_parse():
    for scale_text in const.SCALE_PATTERN:
        for note_text in const.SCALE_PATTERN[scale_text]:
            parser = NoteParser()
            note_data = parser.parse(note_text)

            accidentals_origin = Accidentals(note_text[1:])
            assert accidentals_origin == note_data.accidentals

            note_origin = const.TONE_TO_INDEX[note_text]
            assert note_origin == note_data.note_index


def test_parse_more_than_double_flat():
    parser = NoteParser()

    for i in range(20):
        note_data = parser.parse("C" + (const.ACCIDENTALS_SHARP * i))
        accidentals_origin = Accidentals(const.ACCIDENTALS_SHARP * i)
        assert accidentals_origin == note_data.accidentals

        note_index_origin = i % 12
        assert note_index_origin == note_data.note_index

    for i in range(20):
        note_data = parser.parse("C" + (const.ACCIDENTALS_FLAT * i))
        accidentals_origin = Accidentals(const.ACCIDENTALS_FLAT * i)
        assert accidentals_origin == note_data.accidentals

        note_index_origin = (-i) % 12
        assert note_index_origin == note_data.note_index



from dlchord2 import const
from dlchord2.parser.accidentals_parser import AccidentalsParser, AccidentalsType

accidentals_pattern = {
    "C": (0, AccidentalsType.NONE),
    "Db": (-1, AccidentalsType.FLAT),
    "C#": (1, AccidentalsType.SHARP),
    "D": (0, AccidentalsType.NONE),
    "Eb": (-1, AccidentalsType.FLAT),
    "D#": (1, AccidentalsType.SHARP),
    "E": (0, AccidentalsType.NONE),
    "F": (0, AccidentalsType.NONE),
    "Gb": (-1, AccidentalsType.FLAT),
    "F#": (1, AccidentalsType.SHARP),
    "G": (0, AccidentalsType.NONE),
    "Ab": (-1, AccidentalsType.FLAT),
    "G#": (1, AccidentalsType.SHARP),
    "A": (0, AccidentalsType.NONE),
    "Bb": (-1, AccidentalsType.FLAT),
    "A#": (1, AccidentalsType.SHARP),
    "B": (0, AccidentalsType.NONE),
    "B#": (1, AccidentalsType.SHARP),
    "Dbb": (-2, AccidentalsType.FLAT),
    "C##": (2, AccidentalsType.SHARP),
    "Ebb": (-2, AccidentalsType.FLAT),
    "Fb": (-1, AccidentalsType.FLAT),
    "D##": (2, AccidentalsType.SHARP),
    "E#": (1, AccidentalsType.SHARP),
    "F##": (2, AccidentalsType.SHARP),
    "Abb": (-2, AccidentalsType.FLAT),
    "Bbb": (-2, AccidentalsType.FLAT),
    "G##": (2, AccidentalsType.SHARP),
    "Cb": (-1, AccidentalsType.FLAT)
}


def test_parse():
    for scale_text in const.SCALE_PATTERN:
        for note in const.SCALE_PATTERN[scale_text]:
            parser = AccidentalsParser()
            accidentals_data = parser.parse(note[1:])

            accidentals_origin = note[1:]
            assert accidentals_origin == accidentals_data.accidentals

            accidentals_type_origin = accidentals_pattern[note][1]
            assert accidentals_type_origin == accidentals_data.accidentals_type

            accidentals_trans_num_origin = accidentals_pattern[note][0]
            assert accidentals_trans_num_origin == accidentals_data.transpose_num



from dlchord2 import const
from dlchord2.exceptions.scale_exceptions import ScaleNotFoundError


class Scale(object):
    def __init__(self, scale_text):
        self.__scale_text = scale_text
        if self.__scale_text not in const.SCALE_PATTERN:
            raise ScaleNotFoundError("指定したスケール {} が見つかりません。".format(scale_text))

        self.__scale_list = const.SCALE_PATTERN[self.__scale_text]

    def __str__(self):
        return self.scale_text

    def __unicode__(self):
        return self.scale_text

    def __eq__(self, other):
        if not isinstance(other, Scale):
            raise TypeError("{} オブジェクトとScaleオブジェクトを比較できません".format(type(other)))

        if self.scale_text == other.scale_text and self.scale_list == other.scale_list:
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "<Scale: {}>".format(self.scale_text)

    @property
    def scale_text(self):
        """
        スケールをテキストで取得します。
        :return: スケール
        :rtype: str
        """
        return self.__scale_text

    @property
    def scale_list(self):
        """
        スケールのノートのテキストリストを取得します。
        :return: ノートのテキストリスト
        :rtype: list[str]
        """
        return self.__scale_list

    def get_note_text(self, index):
        """
        指定したインデックスのノートのテキストを取得します。
        インデックスは、現在設定されているスケールから見たインデックスを指します。
        indexが0でCスケールの場合、取得されるノートテキストはCになります。
        indexが0でDbスケールの場合、取得されるノートテキストはDbになります。
        :param index: 現在設定されているスケールから見たインデックス
        :type index: int
        :return: ノートのテキスト
        :rtype: str
        """
        if index < 0 or index > 11:
            raise ScaleNotFoundError("範囲外のインデックスが指定されました。")

        return self.__scale_list[index]

    def get_note_text_absolute(self, absolute_index):
        """
        現在のスケールを移動させ、Cが先頭になるように並びかえられたスケールで、ノートのテキストを取得します。
        indexが0でCスケールの場合、取得されるノートテキストはCになります。
        indexが0でDbスケールの場合、取得されるノートテキストはCになります。
        :param absolute_index:
        :type absolute_index:
        :return: ノートのテキスト
        :rtype: str
        """
        if absolute_index < 0 or absolute_index > 11:
            raise ScaleNotFoundError("範囲外のインデックスが指定されました。")

        index = (absolute_index - const.SCALE_TO_INDEX[self.scale_text]) % 12
        return self.__scale_list[index]



from dlchord2.parser.accidentals_parser import AccidentalsParser


class Accidentals(object):
    def __init__(self, accidentals):
        """
        :param accidentals: ルート音を含まない、調号のみのテキスト
        :type accidentals: str
        """

        parser = AccidentalsParser()
        self.__accidentals_data = parser.parse(accidentals)

    def __eq__(self, other):
        if not isinstance(other, Accidentals):
            raise TypeError("{} オブジェクトとAccidentalsオブジェクトを比較できません".format(type(other)))

        if self.transpose_num == other.transpose_num:
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __unicode__(self):
        return self.text

    def __str__(self):
        return self.text

    def __repr__(self):
        return "<Accidentals: {}>".format(self.text)

    @property
    def text(self):
        """
        調号のテキストを取得します。
        :return: 調号のテキスト
        :rtype: str
        """
        return self.__accidentals_data.accidentals

    @property
    def type(self):
        """
        調号の種類を取得します。
        :return: 調号の種類
        :rtype: AccidentalsType
        """
        return self.__accidentals_data.accidentals_type

    @property
    def transpose_num(self):
        """
        調号の変化量を取得します。
        :return: 調号の変化量
        :rtype: int
        """
        return self.__accidentals_data.transpose_num



from dlchord2 import const
from dlchord2.note import Note
from dlchord2.scale import Scale


def test_scale_text():
    for scale_text in const.SCALE_PATTERN.keys():
        scale = Scale(scale_text)

        assert scale_text == scale.scale_text


def test_scale_list():
    for scale_text in const.SCALE_PATTERN.keys():
        scale = Scale(scale_text)

        assert const.SCALE_PATTERN[scale_text] == scale.scale_list


def test_get_note_text():
    for scale_text in const.SCALE_PATTERN.keys():
        scale = Scale(scale_text)

        for idx in range(12):
            assert const.SCALE_PATTERN[scale_text][idx] == scale.get_note_text(idx)


def test_get_tone_text_absolute():
    for scale_text in const.SCALE_PATTERN.keys():
        scale = Scale(scale_text)

        for idx in range(12):
            note_origin = Note(const.SCALE_PATTERN["C"][idx])
            note_test = Note(scale.get_note_text_absolute(idx))
            assert note_origin == note_test



from dlchord2 import const


class ChordParseData(object):
    def __init__(self, root_text, bass_text, quality_text):
        self.__root_text = root_text
        self.__bass_text = bass_text
        self.__quality_text = quality_text

    @property
    def root_text(self):
        return self.__root_text

    @property
    def bass_text(self):
        return self.__bass_text

    @property
    def quality_text(self):
        return self.__quality_text


class ChordParser(object):
    """
    コードをルート、クオリティ、ベースに分解するクラス。
    """

    def parse(self, chord_text):
        split_text = chord_text.split("/")
        root_text = chord_text[0]
        bass_text = ""
        accidentals_text = ""

        if len(split_text) > 1:
            bass_text = split_text[1]

        res = split_text[0][1:]
        for i in range(len(res)):
            if res[i] != const.ACCIDENTALS_SHARP and res[i] != const.ACCIDENTALS_FLAT:
                break
            accidentals_text += res[i]

        root_text += accidentals_text
        quality_text = res[len(accidentals_text):]

        # ベースがない場合はルート音がベースになります
        if bass_text == "":
            bass_text = root_text

        return ChordParseData(root_text, bass_text, quality_text)




from dlchord2 import const
from dlchord2.accidentals import Accidentals
from dlchord2.exceptions.note_exceptions import NoteParseError


class NoteParseData(object):
    """
    ノート解析データを格納するクラス
    """
    def __init__(self, accidentals, note_index):
        self._accidentals = accidentals
        self._note_index = note_index

    @property
    def accidentals(self):
        """
        ノートの調号を取得します。
        :return: ノートの調号
        :rtype: Accidentals
        """
        return self._accidentals

    @property
    def note_index(self):
        """
        ノートの音階インデックスを取得します。
        このインデックスは調号による変化量を含んでいます。
        :return: ノートの音階インデックス
        :rtype: int
        """
        return self._note_index


class NoteParser(object):
    """
    ノートを解析するクラス
    """

    def parse(self, raw_note_text):
        """
        生のノートテキストから、調号、ルート音を解析します。
        :param raw_note_text: 生のノートテキスト
        :type raw_note_text: str
        :return: ノート解析データ
        :rtype: NoteParseData
        """

        note_text_length = len(raw_note_text)
        if note_text_length == 0:
            raise NoteParseError("ノートのテキストが入力されていません")

        root_note_text = raw_note_text[0]
        if root_note_text not in const.TONE_TO_INDEX:
            raise NoteParseError("存在しないルート音です。")

        note_index = const.TONE_TO_INDEX[root_note_text]
        accidentals = Accidentals(raw_note_text[1:])

        note_index = (note_index + accidentals.transpose_num) % 12
        note_parse_data = NoteParseData(accidentals, note_index)

        return note_parse_data



from dlchord2 import const
from dlchord2.accidentals import Accidentals
from dlchord2.note import Note
from dlchord2.scale import Scale


def test_create_from_index_note():
    for i in range(12):
        note = Note.create_from_index_note(i)


def test_create_from_tension():
    tensions = const.TENSION_TO_INDEX.keys()

    for tension in tensions:
        note = Note.create_from_tension(tension)


def test_transposed():
    for scale_text in const.SCALE_PATTERN.keys():
        scale_origin = Scale(scale_text)

        for note_text in const.SCALE_PATTERN[scale_text]:
            note = Note(note_text, scale_text=scale_text)

            for i in range(12):
                note_transposed = note.transposed(steps=i, scale_text=scale_text)

                note_index_origin = note.note_index
                assert (note_index_origin + i) % 12 == note_transposed.note_index

                assert scale_origin == note_transposed.scale

            for i in range(0, -12, -1):
                note_transposed = note.transposed(steps=i, scale_text=scale_text)

                note_index_origin = note.note_index
                assert (note_index_origin + i) % 12 == note_transposed.note_index

                assert scale_origin == note_transposed.scale


def test_disabled():
    note = Note("C")
    assert not note.disabled

    note.disabled = True
    assert note.disabled


def test_note_index():
    for scale_text in const.SCALE_PATTERN.keys():
        for note_text in const.SCALE_PATTERN[scale_text]:
            note = Note(note_text, scale_text=scale_text)

            note_index_origin = const.TONE_TO_INDEX[note_text]
            assert note_index_origin == note.note_index


def test_raw_note_text():
    for scale_text in const.SCALE_PATTERN.keys():
        for note_text in const.SCALE_PATTERN[scale_text]:
            note = Note(note_text, scale_text=scale_text)

            assert note_text == note.raw_note_text


def test_normed_note_text():
    for scale_text in const.SCALE_PATTERN.keys():
        for note_text in const.SCALE_PATTERN[scale_text]:
            note = Note(note_text, scale_text=scale_text)

            assert note_text == note.normed_note_text


def test_accidentals():
    for scale_text in const.SCALE_PATTERN.keys():
        for note_text in const.SCALE_PATTERN[scale_text]:
            note = Note(note_text, scale_text=scale_text)

            accidentals_origin = Accidentals(note_text[1:])
            assert accidentals_origin == note.accidentals


def test_scale():
    for scale_text in const.SCALE_PATTERN.keys():
        for note_text in const.SCALE_PATTERN[scale_text]:
            note = Note(note_text, scale_text=scale_text)

            scale_origin = Scale(scale_text)
            assert scale_origin == note.scale



from dlchord2.accidentals import Accidentals
from dlchord2 import const
from dlchord2.parser.accidentals_parser import AccidentalsType


def test_text():
    for i in range(20):
        accidentals = Accidentals(const.ACCIDENTALS_SHARP * i)
        accidentals_text_origin = const.ACCIDENTALS_SHARP * i
        assert accidentals_text_origin == accidentals.text

    for i in range(20):
        accidentals = Accidentals(const.ACCIDENTALS_FLAT * i)
        accidentals_text_origin = const.ACCIDENTALS_FLAT * i
        assert accidentals_text_origin == accidentals.text


def test_type():
    accidentals = Accidentals("")
    accidentals_type_origin = AccidentalsType.NONE
    assert accidentals_type_origin == accidentals.type

    for i in range(1, 20):
        accidentals = Accidentals(const.ACCIDENTALS_SHARP * i)
        accidentals_type_origin = AccidentalsType.SHARP
        assert accidentals_type_origin == accidentals.type

    for i in range(1, 20):
        accidentals = Accidentals(const.ACCIDENTALS_FLAT * i)
        accidentals_type_origin = AccidentalsType.FLAT
        assert accidentals_type_origin == accidentals.type


def test_transpose_num():
    accidentals = Accidentals("")
    accidentals_trans_num_origin = 0
    assert accidentals_trans_num_origin == accidentals.transpose_num

    for i in range(1, 20):
        accidentals = Accidentals(const.ACCIDENTALS_SHARP * i)
        accidentals_trans_num_origin = i
        assert accidentals_trans_num_origin == accidentals.transpose_num

    for i in range(1, 20):
        accidentals = Accidentals(const.ACCIDENTALS_FLAT * i)
        accidentals_trans_num_origin = -i
        assert accidentals_trans_num_origin == accidentals.transpose_num

